
from pydantic import BaseConfig
from queenbee.recipe import Recipe

from ._base import APIBase

BaseConfig.allow_population_by_field_name = True


class RecipesAPI(APIBase):

    def get_recipe(self, owner: str, name: str, tag: str = 'latest') -> Recipe:
        res = self.client.get(
            path=f'/registries/{owner}/recipe/{name}/{tag}/json'
        )
        return Recipe.parse_obj(res)

    def add_to_project(self, owner: str, name: str,
                    project_slug: str, tag: str = 'latest') -> str:
        prj_owner, prj_name = project_slug.split('/')
        res = self.client.post(
            path=f'/projects/{prj_owner}/{prj_name}/recipes/filters',
            json={
                "owner": owner,
                "name": name,
                "tag": tag
            }
        )
        return res



import pytest
from pandas.testing import assert_frame_equal
from pollination_io.dataframe import ColumnMeta, RunsDataFrame


@pytest.fixture
def input_columns():
    cols = ['grid-filter', 'model', 'name',
            'radiance-parameters', 'sensor-count',
            'shade_area', 'shade_count', 'shade_depth']
    cols.sort()
    return cols


@pytest.fixture
def output_columns():
    return ['results']


@pytest.fixture
def parameter_columns(input_columns):
    cols = ['grid-filter', 'name',
            'radiance-parameters', 'sensor-count',
            'shade_area', 'shade_count', 'shade_depth']
    cols.sort()
    return cols


@pytest.fixture
def artifact_columns():
    return ['model', 'results']


@pytest.fixture
def meta_columns():
    return ['job-id', 'run-id', 'run-status']


def test_column_meta(run_results_dict, input_columns,
                     output_columns,
                     parameter_columns,
                     artifact_columns,
                     meta_columns,):
    meta = ColumnMeta()
    for r in run_results_dict:
        meta.extend_from_run_result(r)

    assert meta.inputs == input_columns
    assert meta.outputs == output_columns
    assert meta.parameters == parameter_columns
    assert meta.artifacts == artifact_columns
    assert meta.meta == meta_columns


def test_create_df(run_results_dict, input_columns):
    rdf = RunsDataFrame.from_run_results(run_results_dict)

    df = rdf.dataframe

    assert_frame_equal(rdf.inputs, df[input_columns])

    assert_frame_equal(rdf.parameters, df[[
        'grid-filter', 'name',
        'radiance-parameters', 'sensor-count',
        'shade_area', 'shade_count', 'shade_depth'
    ]])
    assert_frame_equal(rdf.artifacts, df[[
        'model', 'results'
    ]])

    assert_frame_equal(rdf.meta, df[['job-id', 'run-status']])
    assert_frame_equal(rdf.outputs, df[['results']])



import typing as t
import numpy as np
import pandas as pd

class ColumnMeta:

    _meta = {'job-id', 'run-id', 'run-status'}

    def __init__(self, inputs: t.List[str] = [], outputs: t.List[str] = [],
                 parameters: t.List[str] = [], artifacts: t.List[str] = []):
        self._inputs = set(inputs)
        self._outputs = set(outputs)
        self._parameters = set(parameters)
        self._artifacts = set(artifacts)

    def extend_from_run_result(self, result: t.List[dict]):
        inputs = []
        outputs = []
        parameters = []
        artifacts = []

        for res in result:
            r_type = res['type'].lower()
            name = res['name']

            if 'input' in r_type:
                if name in self._meta:
                    continue
                inputs.append(name)
            elif 'output' in r_type:
                outputs.append(name)

            if 'file' in r_type or 'folder' in r_type or 'path' in r_type:
                artifacts.append(name)
            else:
                parameters.append(name)

        self._inputs.update(set(inputs))
        self._outputs.update(set(outputs))
        self._parameters.update(set(parameters))
        self._artifacts.update(set(artifacts))

    @staticmethod
    def list(data: set) -> list:
        ls = list(data)
        ls.sort()
        return ls

    @property
    def inputs(self) -> t.List[str]:
        return self.list(self._inputs)

    @property
    def outputs(self) -> t.List[str]:
        return self.list(self._outputs)

    @property
    def meta(self) -> t.List[str]:
        return self.list(self._meta)

    @property
    def parameters(self) -> t.List[str]:
        return self.list(self._parameters)

    @property
    def artifacts(self) -> t.List[str]:
        return self.list(self._artifacts)

    @property
    def input_parameters(self) -> t.List[str]:
        return self.list(self._inputs & self._parameters)

    @property
    def output_parameters(self) -> t.List[str]:
        return self.list(self._outputs & self._parameters)

    @property
    def input_artifacts(self) -> t.List[str]:
        return self.list(self._inputs & self._artifacts)

    @property
    def output_artifacts(self) -> t.List[str]:
        return self.list(self._outputs & self._artifacts)


class RunsDataFrame:

    def __init__(self, df: pd.DataFrame, meta: ColumnMeta = ColumnMeta):
        self._df = df
        self._column_meta = meta

    @classmethod
    def from_run_results(cls, data: t.List[t.List[dict]]) -> 'RunsDataFrame':
        meta = ColumnMeta()
        df = pd.DataFrame()
        for run in data:
            meta.extend_from_run_result(run)
            sub_df = pd.DataFrame(run)
            sub_df = pd.DataFrame(
                {'name': sub_df['name'], 'value': sub_df.apply(cls._get_value, axis=1)})
            job_id = sub_df[sub_df['name'] == 'job-id']['value'].iloc[0]
            run_id = sub_df[sub_df['name'] == 'run-id']['value'].iloc[0]

            sub_df['job_id'] = job_id
            sub_df['run-id'] = run_id
            sub_df.index = [0]*len(sub_df)

            pivot = sub_df.pivot(columns='name', values='value')
            try:
                df = df.append(pivot)
            except AttributeError:
                # Pandas 2.0 - see here: https://stackoverflow.com/a/75956237/4394669
                df = pd.concat([df, pivot])

        df.columns.name = None
        df.set_index('run-id', inplace=True)
        df = df.infer_objects()

        return cls(df=df, meta=meta)

    @staticmethod
    def _get_value(row) -> any:
        if row['value'] is not np.NAN:
            return row['value']
        if row['source'] is not np.NAN:
            # print(row['source'])
            return row['source']['path']
        else:
            raise ValueError('No value to pick from...')

    @property
    def dataframe(self) -> pd.DataFrame:
        return self._df

    @property
    def inputs(self) -> pd.DataFrame:
        return self._df[self._column_meta.inputs]

    @property
    def outputs(self) -> pd.DataFrame:
        return self._df[self._column_meta.outputs]

    @property
    def parameters(self) -> pd.DataFrame:
        return self._df[self._column_meta.parameters]

    @property
    def artifacts(self) -> pd.DataFrame:
        return self._df[self._column_meta.artifacts]

    @property
    def input_parameters(self) -> pd.DataFrame:
        return self._df[self._column_meta.input_parameters]

    @property
    def input_artifacts(self) -> pd.DataFrame:
        return self._df[self._column_meta.input_artifacts]

    @property
    def output_parameters(self) -> pd.DataFrame:
        return self._df[self._column_meta.output_parameters]

    @property
    def output_artifacts(self) -> pd.DataFrame:
        return self._df[self._column_meta.output_artifacts]

    @property
    def meta(self) -> pd.DataFrame:
        return self._df[['job-id', 'run-status']]



import json
from pathlib import Path

import pytest
import requests_mock
from pydantic import BaseConfig
from queenbee.job import Job, JobStatus, RunStatus
from queenbee.recipe import RecipeInterface
from queenbee.recipe.recipe import Recipe
from requests import Request

BaseConfig.allow_population_by_field_name = True


@pytest.fixture
def curdir():
    return Path(__file__).parent


@pytest.fixture
def assets(curdir: Path):
    return curdir.joinpath('assets')


@pytest.fixture
def run_results_dict(assets: Path):
    with open(assets.joinpath('run-results.json')) as f:
        return json.load(f)


@pytest.fixture
def artifact_list(assets: Path):
    with open(assets.joinpath('artifact-list.json')) as f:
        return json.load(f)


@pytest.fixture
def single_job(assets: Path):
    return json.loads(assets.joinpath('job-pollination-api.json').read_bytes())


@pytest.fixture
def single_run(assets: Path):
    return json.loads(assets.joinpath('run-pollination-api.json').read_bytes())


@pytest.fixture
def recipe(assets: Path):
    return Recipe.from_file(assets.joinpath('recipe-pollination-api.json').as_posix())


@pytest.fixture
def default_host():
    return 'https://api.pollination.cloud'


@pytest.fixture
def custom_host():
    return 'https://api.staging.pollination.cloud'


@pytest.fixture
def api_token():
    return 'some-long-token-string'


@pytest.fixture
def job_id(single_job):
    return single_job['id']


@pytest.fixture
def run_id(single_run):
    return single_run['id']


@pytest.fixture
def job_spec(single_job):
    return Job.parse_obj(single_job['spec'])


@pytest.fixture
def job_status(single_job):
    return JobStatus.parse_obj(single_job['status'])


@pytest.fixture
def recipe_interface(single_job):
    return RecipeInterface.parse_obj(single_job['recipe'])


@pytest.fixture
def run_status(single_run):
    return RunStatus.parse_obj(single_run['status'])


@pytest.fixture
def job_create_response(job_id):
    return {
        "id": job_id,
        "message": "Use Location in headers to access the new object."
    }


@pytest.fixture
def run_output():
    return 'results'


@pytest.fixture
def artifact_path():
    return 'some-path'


@pytest.fixture
def artifact_url():
    return 'https://gcs.some.random.url/artifat-path'


@pytest.fixture
def job_results(default_host, job_id, run_results_dict):
    with requests_mock.Mocker() as m:
        m.get(
            f'{default_host}/projects/ladybug-tools/demo/results?job_id={job_id}&page=1',
            json={'resources': run_results_dict}
        )
        yield


@pytest.fixture
def job_artifacts_list(default_host, job_id, artifact_path, artifact_list):
    with requests_mock.Mocker() as m:
        m.get(
            f'{default_host}/projects/ladybug-tools/demo/jobs/{job_id}/artifacts?path={artifact_path}',
            json=artifact_list
        )
        yield


@pytest.fixture
def download_job_artifact(default_host, job_id, artifact_path, artifact_url):
    with requests_mock.Mocker() as m:
        m.get(
            f'{default_host}/projects/ladybug-tools/demo/jobs/{job_id}/artifacts/download?path={artifact_path}',
            text=artifact_url,
        )
        m.get(artifact_url, content=b'Hello World!')
        yield


@pytest.fixture
def download_run_output_artifact(default_host, run_id, run_output, artifact_url):
    with requests_mock.Mocker() as m:
        m.get(
            f'{default_host}/projects/ladybug-tools/demo/runs/{run_id}/outputs/{run_output}',
            text=artifact_url,
        )
        m.get(artifact_url, content=b'Hello World!')
        yield


@pytest.fixture
def single_run_from_api(default_host, run_id, single_run):
    with requests_mock.Mocker() as m:
        m.get(
            f'{default_host}/projects/ladybug-tools/demo/runs/{run_id}',
            json=single_run,
        )
        yield


@pytest.fixture
def single_job_from_api(default_host, job_id, single_job):
    with requests_mock.Mocker() as m:
        m.get(
            f'{default_host}/projects/ladybug-tools/demo/jobs/{job_id}',
            json=single_job,
        )
        yield


@pytest.fixture
def create_job(default_host, job_spec, job_create_response):
    def additional_matcher(request: Request):
        return json.loads(request.text) == job_spec.dict()

    with requests_mock.Mocker() as m:
        m.post(
            f'{default_host}/projects/ladybug-tools/demo/jobs',
            additional_matcher=additional_matcher,
            json=job_create_response
        )
        yield


@pytest.fixture
def get_recipe(default_host, recipe):
    with requests_mock.Mocker() as m:
        m.get(
            f'{default_host}/registries/ladybug-tools/recipe/annual-daylight/0.8.2-viz/json',
            json=recipe.dict(),
        )
        yield


@pytest.fixture
def get_recipe_and_create_job(default_host, recipe, job_spec, job_create_response):
    def additional_matcher(request: Request):
        return json.loads(request.text) == job_spec.dict()

    with requests_mock.Mocker() as m:
        m.get(
            f'{default_host}/registries/ladybug-tools/recipe/annual-daylight/0.8.2-viz/json',
            json=recipe.dict(),
        )
        m.post(
            f'{default_host}/projects/ladybug-tools/demo/jobs',
            additional_matcher=additional_matcher,
            json=job_create_response
        )
        yield


@pytest.fixture
def user_profile():
    return {
        "id": "96c12d05-f1a2-4491-b0cc-c2ed473301b5",
        "email": "ladybugbot@ladybug.tools",
        "name": "Ladybug Bot",
        "username": "ladybugbot",
        "description": "Beep Boop!",
        "picture": "https://avatars1.githubusercontent.com/u/38131342"
    }


@pytest.fixture
def get_user(default_host, user_profile):
    with requests_mock.Mocker() as m:
        m.get('/user', json=user_profile)
        yield



from uuid import uuid4

import pytest
from pandas.testing import assert_frame_equal
from pollination_io.api.client import ApiClient
from pollination_io.dataframe import RunsDataFrame
from pollination_io.interactors import (Artifact, Job, NewJob, Recipe,
                                               Run)


@pytest.fixture
def owner():
    return 'ladybug-tools'


@pytest.fixture
def project():
    return 'demo'


@pytest.fixture
def job_id():
    return str(uuid4())


@pytest.fixture
def test_client():
    pass


def test_artifact():
    key = '/some/test/key.txt'
    file_type = 'file'

    class JobStub:
        @staticmethod
        def download_artifact(path: str) -> bool:
            return path == key

    art = Artifact(key, file_type, JobStub)
    assert art.name == 'key.txt'
    assert art.is_file
    assert not art.is_folder
    assert art.download()

    art = Artifact(key, 'folder', JobStub)
    assert art.name == 'key.txt'
    assert not art.is_file
    assert art.is_folder
    with pytest.raises(ValueError):
        art.download()


def test_job(job_id, single_job_from_api, job_spec, job_status, recipe_interface):
    job = Job('ladybug-tools', 'demo', job_id)

    assert job.spec == job_spec
    assert job.status == job_status
    assert job.recipe_interface == recipe_interface
    assert job.recipe.owner == 'ladybug-tools'
    assert job.recipe.name == 'annual-daylight'
    assert job.recipe.tag == '0.8.2-viz'


def test_job_runs(job_id, job_results, run_results_dict):
    job = Job('ladybug-tools', 'demo', job_id)
    rdf = RunsDataFrame.from_run_results(run_results_dict)

    assert_frame_equal(job.runs_dataframe.dataframe, rdf.dataframe)

    run_id = job.runs_dataframe.dataframe.index[0]

    run = job.runs[0]
    assert run.owner == 'ladybug-tools'
    assert run.project == 'demo'
    assert run.job_id == job_id
    assert run.id == run_id


def test_run(job_id, run_id, single_run_from_api, run_status, recipe_interface):
    run = Run('ladybug-tools', 'demo', job_id, run_id)
    assert run.status == run_status
    assert run.recipe_interface == recipe_interface
    assert run.recipe.owner == 'ladybug-tools'
    assert run.recipe.name == 'annual-daylight'
    assert run.recipe.tag == '0.8.2-viz'

    assert run.full_artifact_path(
        '/some/path') == f'runs/{run_id}/workspace/some/path'
    assert run.full_artifact_path(
        'some/path') == f'runs/{run_id}/workspace/some/path'


def test_recipe(get_recipe, recipe):
    r = Recipe('ladybug-tools', 'annual-daylight', '0.8.2-viz')

    assert r.source_url == 'https://api.pollination.cloud/registries/ladybug-tools/recipe/annual-daylight/0.8.2-viz'
    assert r.api_object == recipe
    assert r.input_artifacts == ['model', 'schedule', 'wea']
    assert r.input_parameters == ['cpu-count', 'grid-filter',
                                  'min-sensor-count', 'north',
                                  'radiance-parameters', 'thresholds']
    assert r.inputs_required == ['model', 'wea']


def test_new_job_validation(get_recipe):
    r = Recipe('ladybug-tools', 'annual-daylight', '0.8.2-viz')

    new_job = NewJob('ladybug-tools', 'demo', r)

    with pytest.raises(ValueError):
        new_job.create()

    new_job.arguments = [
        {'model': 'path/to/model', 'wea': 'path/to/wea'},
        {'model': 'path/to/model'}
    ]

    with pytest.raises(ValueError):
        new_job.create()


def test_new_job(get_recipe_and_create_job, job_id):
    r = Recipe('ladybug-tools', 'annual-daylight', '0.8.2-viz')
    new_job = NewJob(
        owner='ladybug-tools',
        project='demo',
        recipe=r,
        name='Annual daylight with visualization',
        description='Show case the new visualization output',
        arguments=[
            {
                'cpu-count': '50',
                'grid-filter': '*',
                'min-sensor-count': '500',
                'north': '0',
                'radiance-parameters': '-ab 2 -ad 5000 -lw 2e-05',
                'thresholds': '-t 300 -lt 100 -ut 3000',
                'wea': 'USA_CA_Palo.Alto.AP.724937_CTZ2010.wea',
                'model': 'hsog4fmk.jjv.hbjson'
            },
            {
                'cpu-count': '50',
                'grid-filter': '*',
                'min-sensor-count': '500',
                'north': '0',
                'radiance-parameters': '-ab 2 -ad 5000 -lw 2e-05',
                'thresholds': '-t 300 -lt 100 -ut 3000',
                'wea': 'USA_CA_Palo.Alto.AP.724937_CTZ2010.wea',
                'model': 'hsog4fmk.jjv.hbjson'
            }
        ]
    )

    job = new_job.create()
    assert isinstance(job, Job)
    assert job.id == job_id
    assert job.owner == new_job.owner
    assert job.project == new_job.project



import pathlib
import typing as t
from io import BytesIO
from uuid import UUID

from queenbee.io.inputs.job import JobArgument, JobPathArgument
from queenbee.job import Job as QbJob
from queenbee.job import JobArguments, JobStatus, RunStatus
from queenbee.recipe import Recipe as QbRecipe
from queenbee.recipe import RecipeInterface

from pollination_io.api.user import UserApi

from .api.client import ApiClient
from .api.jobs import JobsAPI
from .api.recipes import RecipesAPI
from .api.runs import RunsAPI
from .dataframe import RunsDataFrame


class Recipe:

    def __init__(self, owner: str,
                 name: str,
                 tag: str = 'latest',
                 client: ApiClient = ApiClient()):
        self.owner = owner
        self.name = name
        self.tag = tag

        self.recipe_api = RecipesAPI(client)
        self._client = client

        self._api_object = None

    @classmethod
    def from_source_url(cls, url: str, client: ApiClient = ApiClient()):
        source_split = url.split('/')
        return cls(
            owner=source_split[-4],
            name=source_split[-2],
            tag=source_split[-1],
            client=client,
        )

    def _fetch_recipe(self):
        self._api_object = self.recipe_api.get_recipe(
            self.owner, self.name, self.tag)

    def add_to_project(self, project_slug: str):
        res = self.recipe_api.add_to_project(
            self.owner,
            self.name,
            project_slug,
            self.tag)
        return res

    @property
    def source_url(self) -> str:
        return f'{self._client.host}/registries/{self.owner}/recipe/{self.name}/{self.tag}'

    @property
    def api_object(self) -> QbRecipe:
        if self._api_object is None:
            self._fetch_recipe()
        return self._api_object

    @property
    def input_artifacts(self) -> t.List[str]:
        return [i.name for i in self.api_object.inputs if i.is_artifact]

    @property
    def input_parameters(self) -> t.List[str]:
        return [i.name for i in self.api_object.inputs if i.is_parameter]

    @property
    def inputs_required(self) -> t.List[str]:
        return [i.name for i in self.api_object.inputs if i.required]


class Job:

    def __init__(self, owner: str, project: str, id: str, client: ApiClient = ApiClient()):
        self.owner = owner
        self.project = project
        self.id = id
        self.job_api = JobsAPI(client)
        self.run_api = RunsAPI(client)

        self._client = client

        self._runs = None
        self._api_object = None

    def __str__(self) -> str:
        return f'<Job {self.owner}/{self.project}/{self.id}>'

    def _fetch_runs(self):
        self._runs = self.run_api.get_runs(self.owner, self.project, self.id)

    def _fetch_job(self):
        self._api_object = self.job_api.get_job(
            self.owner, self.project, self.id)

    def refresh(self):
        self._fetch_runs()
        self._fetch_job()

    @property
    def api_object(self) -> t.Dict[str, t.Any]:
        if self._api_object is None:
            self._fetch_job()
        return self._api_object

    @property
    def spec(self) -> QbJob:
        return QbJob.parse_obj(self.api_object['spec'])

    @property
    def status(self) -> JobStatus:
        return JobStatus.parse_obj(self.api_object['status'])

    @property
    def recipe_interface(self) -> RecipeInterface:
        return RecipeInterface.parse_obj(self.api_object['recipe'])

    @property
    def recipe(self) -> Recipe:
        return Recipe.from_source_url(self.recipe_interface.source, self._client)

    @property
    def runs_dataframe(self) -> RunsDataFrame:
        if self._runs is None:
            self._fetch_runs()
        return RunsDataFrame.from_run_results(self._runs)

    @property
    def runs(self) -> t.List['Run']:
        df = self.runs_dataframe
        runs = []
        for run_id in df.dataframe.index:
            run = Run(
                self.owner, self.project, self.id, run_id, self._client
            )
            runs.append(run)

        return runs

    def list_artifacts(
            self, path: str = None, page: int = 1,
            per_page: int = 25) -> t.List['Artifact']:
        response = self.job_api.list_job_artifacts(
            self.owner, self.project, self.id, path, page, per_page
        )
        return [
            Artifact(job=self, **artifact)
            for artifact in response['resources']
        ]

    def download_artifact(self, path: str) -> BytesIO:
        return self.job_api.get_job_artifact(self.owner, self.project, self.id, path)


class NewJob:

    def __init__(self, owner: str, project: str, recipe: Recipe,
                 arguments: t.List[t.Dict[str, t.Any]] = [],
                 name: str = None, description: str = None,
                 client: ApiClient = ApiClient()):
        self.owner = owner
        self.project = project
        self.recipe = recipe
        self.arguments = arguments
        self.name = name
        self.description = description

        self._client = client
        self.job_api = JobsAPI(client)

    def create(self) -> Job:
        qb_job = self.generate_qb_job()
        job_id = self.job_api.create_job(self.owner, self.project, qb_job)
        return Job(self.owner, self.project, job_id, self._client)

    def upload_artifact(self, fp: pathlib.Path, target_folder: str = '') -> str:
        artifact_path = pathlib.Path(target_folder).joinpath(fp.name)
        return self.job_api.upload_artifact(self.owner, self.project, fp, artifact_path)

    def generate_qb_job(self) -> QbJob:
        arguments = self._generate_qb_job_arguments()
        return QbJob(
            source=self.recipe.source_url,
            arguments=arguments,
            name=self.name,
            description=self.description,
        )

    def _check_arguments(self):
        if self.arguments == []:
            raise ValueError('No job arguments specified')
        for i, run_args in enumerate(self.arguments):
            for ri in self.recipe.inputs_required:
                if ri not in run_args.keys():
                    raise ValueError(
                        f'Missing required input {ri} in arguments[{i}]')

    def _generate_qb_job_arguments(self) -> t.List[t.List[JobArguments]]:
        self._check_arguments()

        job_arguments = []

        for args in self.arguments:
            run_args = []
            for k, v in args.items():
                if k in self.recipe.input_artifacts:
                    run_args.append(JobPathArgument.parse_obj({
                        'name': k,
                        'source': {
                            'type': 'ProjectFolder',
                            'path': v
                        }
                    }))
                else:
                    run_args.append(JobArgument(
                        name=k,
                        value=v
                    ))

            job_arguments.append(run_args)

        return job_arguments


class Run:

    def __init__(self, owner: str, project: str, job_id: str, id: str, client: ApiClient = ApiClient()) -> None:
        self.owner = owner
        self.project = project
        self.job_id = job_id
        self.id = id

        self.run_api = RunsAPI(client)
        self.job = Job(owner, project, job_id, client)

        self._client = client

        self._api_object = None

    @property
    def _base_artifact_path(self) -> str:
        return f'runs/{self.id}/workspace/'

    def _fetch_run(self):
        self._api_object = self.run_api.get_run(
            self.owner, self.project, self.id)

    def refresh(self):
        self._fetch_run()

    @property
    def api_object(self) -> t.Dict[str, t.Any]:
        if self._api_object is None:
            self._fetch_run()
        return self._api_object

    @property
    def status(self) -> RunStatus:
        return RunStatus.parse_obj(self.api_object['status'])

    @property
    def recipe_interface(self) -> RecipeInterface:
        return RecipeInterface.parse_obj(self.api_object['recipe'])

    @property
    def recipe(self) -> Recipe:
        return Recipe.from_source_url(self.recipe_interface.source, self._client)

    def full_artifact_path(self, path: str) -> str:
        if path.startswith(self._base_artifact_path):
            return path
        if path.startswith('/'):
            path = path[1:]
        return self._base_artifact_path + path

    def download_zipped_output(self, output_name: str) -> BytesIO:
        return self.run_api.download_zipped_run_output(
            self.owner, self.project, self.id, output_name
        )

    def list_artifacts(self, path: str) -> t.List['Artifact']:
        return self.job.list_artifacts(self.full_artifact_path(path))

    def download_artifact(self, path: str) -> BytesIO:
        return self.job.download_artifact(self.full_artifact_path(path))


class Artifact:

    def __init__(self, key: str, file_type: str, job: Job, **kwargs):
        self.key = key
        self.file_type = file_type
        self.job = job

    def __str__(self) -> str:
        return f'<Artifact: {self.file_type} - {self.key}>'

    def __repr__(self) -> str:
        return self.__str__()

    @property
    def name(self) -> str:
        return self.key.split('/')[-1]

    @property
    def is_file(self) -> bool:
        return self.file_type == 'file'

    @property
    def is_folder(self) -> bool:
        return self.file_type == 'folder'

    def download(self) -> BytesIO:
        if self.is_file:
            return self.job.download_artifact(self.key)
        raise ValueError(f'Cannot download artifact of type {self.file_type}')

    def list_children(self) -> t.List['Artifact']:
        return self.job.list_artifacts(self.key)


class AuthUser:

    def __init__(self, client: ApiClient) -> None:
        self.user_api = UserApi(client)
        self._client = client
        self._api_object = None

    def __str__(self) -> str:
        return f'<AuthUser: {self.username}>'

    def __repr__(self) -> str:
        return self.__str__()

    @property
    def api_object(self) -> t.Dict[str, t.Any]:
        if self._api_object is None:
            self._api_object = self.user_api.get_user()
        return self._api_object

    @property
    def id(self) -> UUID:
        return UUID(self.api_object['id'])

    @property
    def name(self) -> str:
        return self.api_object['name']

    @property
    def username(self) -> str:
        return self.api_object['username']

    @property
    def description(self) -> str:
        return self.api_object['description']

    @property
    def picture(self) -> str:
        return self.api_object['picture']



import pytest
from pollination_io.api.client import ApiClient
from pollination_io.api.jobs import JobsAPI
from pollination_io.api.recipes import RecipesAPI
from pollination_io.api.runs import RunsAPI
from pollination_io.api.user import UserApi
from queenbee.job import Job as QbJob


@pytest.fixture
def test_client():
    return ApiClient()


@pytest.fixture
def runs_api(test_client):
    return RunsAPI(test_client)


@pytest.fixture
def jobs_api(test_client):
    return JobsAPI(test_client)


@pytest.fixture
def recipes_api(test_client):
    return RecipesAPI(test_client)


@pytest.fixture
def user_api(test_client):
    return UserApi(test_client)


def test_init_api_client(default_host, custom_host, api_token):
    client_no_auth = ApiClient()
    assert client_no_auth.host == default_host
    assert client_no_auth.headers == {}

    client_with_auth = ApiClient(host=custom_host, api_token=api_token)
    assert client_with_auth.host == custom_host
    assert client_with_auth.headers == {
        'x-pollination-token': api_token
    }

    client_host_string = ApiClient(host='https://some.host/')
    assert client_host_string.host == 'https://some.host'


def test_download_artifact(test_client: ApiClient, download_job_artifact, artifact_url: str):
    res = test_client.download_artifact(artifact_url)
    assert res.read() == b'Hello World!'


def test_get_job(jobs_api: JobsAPI, single_job, single_job_from_api, job_id):
    assert jobs_api.get_job('ladybug-tools', 'demo',
                            job_id) == single_job


def test_get_runs(runs_api: RunsAPI, job_results, job_id, run_results_dict):
    assert runs_api.get_runs('ladybug-tools', 'demo',
                             job_id) == run_results_dict


def test_get_run(runs_api: RunsAPI, single_run, single_run_from_api, run_id):
    assert runs_api.get_run('ladybug-tools', 'demo',
                            run_id) == single_run


def test_list_job_artifacts(jobs_api: JobsAPI, job_artifacts_list, job_id, artifact_path, artifact_list):
    assert jobs_api.list_job_artifacts(
        'ladybug-tools', 'demo', job_id, artifact_path) == artifact_list


def test_get_job_artifact(jobs_api: JobsAPI, download_job_artifact, job_id, artifact_path):
    assert jobs_api.get_job_artifact(
        'ladybug-tools', 'demo', job_id, artifact_path).read() == b'Hello World!'


def test_download_zipped_run_output(runs_api: RunsAPI, download_run_output_artifact, run_id, run_output):
    assert runs_api.download_zipped_run_output(
        'ladybug-tools', 'demo', run_id, run_output).read() == b'Hello World!'


def test_create_job(jobs_api: JobsAPI, create_job, job_id, job_spec: QbJob):
    assert jobs_api.create_job('ladybug-tools', 'demo', job_spec) \
        == job_id


def test_get_recipe(recipes_api: RecipesAPI, get_recipe, recipe):
    assert recipes_api.get_recipe(
        'ladybug-tools', 'annual-daylight', '0.8.2-viz') == recipe


def test_get_auth_user(user_api: UserApi, get_user, user_profile):
    assert user_api.get_user() == user_profile


@pytest.mark.parametrize('api_token,jwt,expected', [
    ('', '', False),
    ('foo', None, True),
    (None, 'jwt-token-string', True),
    (None, '', False),
    (None, None, False),
])
def test_is_authenticated(test_client: ApiClient, api_token, jwt, expected):
    test_client.api_token = api_token
    test_client.jwt_token = jwt
    assert test_client.is_authenticated == expected



import typing as t
from io import BytesIO

from ._base import APIBase


class RunsAPI(APIBase):

    def _run_results_request(self, owner: str, project: str, job_id: str, page: int = 1) -> t.List[t.List[dict]]:
        res = self.client.get(
            path=f'/projects/{owner}/{project}/results',
            params={'job_id': job_id, 'page': page}
        )
        data = res.get('resources', {})
        next_page = res.get('next_page')
        if next_page is not None:
            data.extend(self._run_results_request(
                owner, project, job_id, next_page))
        return data

    def get_runs(self, owner: str, project: str, job_id: str) -> t.List[t.List[dict]]:
        return self._run_results_request(owner, project, job_id)

    def get_run(self, owner: str, project: str, run_id: str) -> t.List[dict]:
        return self.client.get(
            path=f'/projects/{owner}/{project}/runs/{run_id}'
        )

    def download_zipped_run_output(self, owner: str, project: str, run_id: str, output_name: str) -> BytesIO:
        signed_url = self.client.get(
            path=f'/projects/{owner}/{project}/runs/{run_id}/outputs/{output_name}',
        )
        return self.client.download_artifact(signed_url)



import pathlib
import typing as t
from io import BytesIO

from queenbee.job import Job as QbJob

from ._base import APIBase


class JobsAPI(APIBase):

    def upload_artifact(self, owner: str, project: str, filepath: pathlib.Path, artifactpath: pathlib.Path) -> str:
        fp_string = filepath.as_posix()
        ap_string = artifactpath.as_posix()

        res = self.client.post(
            f'/projects/{owner}/{project}/artifacts',
            json={'key': ap_string}
        )

        files = {'file': (fp_string, open(fp_string, 'rb'))}
        res = self.client.session.post(
            url=res['url'], data=res['fields'], files=files)
        res.raise_for_status()

        return ap_string

    def create_job(self, owner: str, project: str, job: QbJob) -> str:
        res = self.client.post(
            path=f'/projects/{owner}/{project}/jobs',
            json=job.dict()
        )
        return res['id']

    def get_job(self, owner: str, project: str, job_id: str) -> t.List[dict]:
        return self.client.get(
            path=f'/projects/{owner}/{project}/jobs/{job_id}'
        )

    def get_job_artifact(self, owner: str, project: str, job_id: str, path: str) -> BytesIO:
        signed_url = self.client.get(
            path=f'/projects/{owner}/{project}/jobs/{job_id}/artifacts/download',
            params={'path': path}
        )
        return self.client.download_artifact(signed_url)

    def list_job_artifacts(
        self, owner: str, project: str, job_id: str, path: str = None,
        page: int = 1, per_page: int = 25
            ) -> t.List[t.Dict[str, t.Any]]:
        return self.client.get(
            path=f'/projects/{owner}/{project}/jobs/{job_id}/artifacts',
            params={'path': path, 'per_page': per_page, 'page': page}
        )



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

with open('requirements.txt') as f:
    requirements = f.read().splitlines()

setuptools.setup(
    name="pollination-io",
    use_scm_version=True,
    setup_requires=['setuptools_scm'],
    author="Pollination",
    author_email="info@pollination.cloud",
    description="A Pollination extension to facilitate working with Pollination SDK.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/pollination/pollination-io",
    packages=setuptools.find_packages(exclude=["tests"]),
    install_requires=requirements,
    classifiers=[
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: Implementation :: CPython",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent"
    ],
    license="Apache-2.0 License"
)



import os
import typing as t
from io import BytesIO

import requests

DEFAULT_HOST = os.getenv('POLLINATION_API_URL',
                         'https://api.pollination.cloud')


class ApiClient():

    def __init__(self, host: str = DEFAULT_HOST, api_token: str = None, jwt_token: str = None):
        if host[-1] == '/':
            host = host[:-1]

        self._host = host
        self.api_token = api_token
        self.jwt_token = jwt_token

    @property
    def api_token(self) -> str:
        return self._api_token

    @api_token.setter
    def api_token(self, value):
        self._api_token = self._validate_string_input(value)

    @property
    def jwt_token(self) -> str:
        return self._jwt_token

    @jwt_token.setter
    def jwt_token(self, value):
        self._jwt_token = self._validate_string_input(value)

    @staticmethod
    def _validate_string_input(value) -> str:
        if isinstance(value, str):
            if len(value) == 0:
                return None
            else:
                return value
        if value is None:
            return value
        raise ValueError(
            f'Expected instance of type str but got: {type(value)}')

    @property
    def is_authenticated(self) -> bool:
        return self.api_token is not None or self.jwt_token is not None

    @property
    def host(self) -> str:
        return self._host

    @property
    def headers(self):
        if self.api_token is not None:
            return {
                'x-pollination-token': self.api_token
            }
        elif self.jwt_token is not None:
            return {
                'Authorization': f'Bearer {self.jwt_token}'
            }
        return {}

    @property
    def session(self):
        s = requests.Session()
        s.headers = self.headers
        return s

    def _url_path(self, path: str) -> str:
        if not path.startswith('/'):
            path = '/' + path
        return self.host + path

    def get(self, path: str, params: t.Dict[str, t.Any] = {}) -> t.Dict[str, t.Any]:
        res = self.session.get(url=self._url_path(path), params=params)
        res.raise_for_status()
        try:
            return res.json()
        except:
            return res.text

    def post(self, path: str, json: t.Dict[str, t.Any]) -> t.Dict[str, t.Any]:
        res = self.session.post(url=self._url_path(path), json=json)
        res.raise_for_status()
        try:
            return res.json()
        except:
            return res.text

    def download_artifact(self, signed_url: str) -> BytesIO:
        res = requests.get(signed_url)
        res.raise_for_status()
        return BytesIO(res.content)
